{------------------------------------------------------------------------------}
{   MCM Design                                                                 }
{   Marc Christopher Martin                                                    }
{   Australiensvej 34                                                          }
{   2100 Copenhagen OE                                                         }
{   Denmark                                                                    }
{                                                                              }
{   e-mail support@mcm-design.dk                                               }
{                                                                              }
{------------------------------------------------------------------------------}

unit Usrctwn;

interface

uses
  WinTypes, WinProcs, SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Menus, Printers,
  {$IFNDEF VER90} { JPEG is not included in Delphi 2. }
    JPEG,
  {$ENDIF}
  USAbout,
  uIOFile,                              { MCM units.                 }
  twain, twd_type, twd_prot, twd_Caps;  { Twain Units.               }


const TwainApp = 'TWDSRC32'; { TWAIN drivers "application name".               }

type
  TFormSource = class(TForm)
    OpenDialog   : TOpenDialog;
    SaveDialog   : TSaveDialog;
    SetupDialog  : TPrinterSetupDialog;
    PrintDialog  : TPrintDialog;
    MainMenu1    : TMainMenu;
    FileMenu     : TMenuItem;
    FileNew      : TMenuItem;
    FileOpen     : TMenuItem;
    FileSave     : TMenuItem;
    N1           : TMenuItem;
    FileSnap     : TMenuItem;
    FileGrap     : TMenuItem;
    N2           : TMenuItem;
    PrintItem    : TMenuItem;
    SetupPrinter : TMenuItem;
    N3           : TMenuItem;
    FileExit     : TMenuItem;

    ImageMenu    : TMenuItem;
    GainOffset   : TMenuItem;
    Invert       : TMenuItem;
    Resolution   : TMenuItem;

    HelpMenu     : TMenuItem;
    About        : TMenuItem;

    procedure FormCreate       (    Sender : TObject);
    procedure FormClose        (    Sender : TObject;
                                var Action : TCloseAction);
    procedure FormDestroy      (    Sender : TObject);
    procedure FormPaint        (    Sender : TObject);
    procedure UpdateMenuItems  (    Sender : TObject);

    procedure FileNewClick     (    Sender : TObject);
    procedure FileOpenClick    (    Sender : TObject);
    procedure FileSaveClick    (    Sender : TObject);
    procedure FileSnapClick    (    Sender : TObject);
    procedure FileGrapClick    (    Sender : TObject);
    procedure PrintItemClick   (    Sender : TObject);
    procedure SetupPrinterClick(    Sender : TObject);
    procedure FileExitClick    (    Sender : TObject);
    procedure GainOffsetClick  (    Sender : TObject);
    procedure InvertClick      (    Sender : TObject);
    procedure ResolutionClick  (    Sender : TObject);
    procedure AboutClick       (    Sender : TObject);
  private
    { Private declarations }
    procedure DoSendExit(var Msg    : TMessage);   message 1999;
  public
    { Public declarations }
    IsDeviceOK     : boolean;       { Is Hardware Device OK.                   }
    IsDirty        : boolean;       { Does the TWAIN interface hold an image.  }
    IsNewFile      : boolean;       {                                          }
    IsSaved        : boolean;       { Is image saved ?                         }
    Bitmap         : TBitmap;       { The Bitmap contains the image.           }
    ImageStream    : TMemoryStream; { Memory stream used to hold DIB.          }
    pBmpInfo       : PBitmapInfo;   { Pointer to Bitmap Info in ImageStream.   }
    pBmp           : Pointer;       { Pointer to Bitmap data in ImageStream.   }
    {$IFNDEF VER90}
      JPEGImage    : TJPEGImage;
    {$ENDIF}

    function  CopyBmpToStream : boolean; { Makes a copy of bitmap in Stream.   }
    { IMPORTANT !!!!                                                           }
    { Assign function below to Bitmap.OnChange, whenever Bitmap is Created.    }
    { This ensures that functions elsewhere relying on pBmp & pBmpInfo, auto-  }
    { matically copy bitmap onto the streamer when needed, but only when the   }
    { bitmap has changed format (contained image).                             }
    procedure ImageChanged(Sender : TObject);
    procedure AutoOpenPic;
  end;

var FormSource : TFormSource;


implementation

{$R *.DFM}

procedure TFormSource.FormCreate(Sender : TObject);
begin
  Screen.OnActiveFormChange := UpdateMenuItems;

  ClientWidth  := 512; { Any default client Width.                             }
  ClientHeight := 512; { Any default client Height.                            }

  IsNewFile    := True;
  IsDirty      := False;
  IsSaved      := True;

  { Set up bitmap handles.                                                     }
  pBmp            := Nil;
  pBmpInfo        := Nil;
  Bitmap          := TBitmap.Create;
  Bitmap.OnChange := ImageChanged;
  ImageStream     := TMemoryStream.Create;

  {---- Check Device Hardware. ----}
  { Insert Your code.              }
  { Check if device is present.    }
  if (True) { Is_Device_Present ?  }
  then begin
       IsDeviceOK := True;

       {---- Initialize Device Hardware. ----}
       { Insert Your code.                   }

  end
  else begin { Hardware was not present or functioning correctly !             }
       messagebeep(0);
       IsDeviceOK := False;
       Messagebox(Handle, 'Return to App.', 'Device Not Found !', Mb_OK);
       NotifyCloseDSReq;
  end;
end; { End TFormSource.FormCreate.                                             }


procedure TFormSource.AutoOpenPic;
begin
  pBmp     := Nil;
  pBmpInfo := Nil;
  Bitmap.Free;
  Bitmap := TBitmap.Create;
  Bitmap.OnChange := ImageChanged;

  { Acquire image automatically.                     }
  { ACQUIRE/LOAD YOUR IMAGE INTO BITMAP HERE !!!!!!! }
  FileOpenClick(Self);
end; { End TFormSource.AutoOpenPic.                                            }


procedure TFormSource.FormClose(Sender : TObject; var Action : TCloseAction);
begin
  if IsDeviceOK
  then begin
  {---- Stop image acquisition. ----}
  { Insert Your code.               }

  end;
  { Should not return an image.                                                }
  NotifyCloseDSReq;
end; { End TFormSource.FormClose.                                              }


procedure TFormSource.FormDestroy(Sender : TObject);
begin
  { Set Screen.OnActiveFormChange to Nil to avoid access vialation.            }
  Screen.OnActiveFormChange := Nil;

  { Clear Bitmap, if any.                                                      }
  Bitmap.Free;
  ImageStream.Free;
end; { End TFormSource.FormDestroy.                                            }


procedure TFormSource.DoSendExit(var Msg : TMessage);
begin
  FormSource.FileExitClick(FormSource)
end; { End TFormSource.DoSendExit.                                             }


procedure TFormSource.FormPaint(Sender : TObject);
begin
  if Not(Bitmap.Empty)
  then Canvas.Draw(0, 0, Bitmap);
end; { End TFormSource.FormPaint.                                              }


procedure TFormSource.UpdateMenuItems(Sender : TObject);
var EnableMenu : boolean;
begin
  EnableMenu         := Not(Bitmap.Empty);
  FileSave.Enabled   := EnableMenu;
  PrintItem.Enabled  := EnableMenu;

  GainOffset.Enabled := EnableMenu;
  Invert.Enabled     := EnableMenu;
  Resolution.Enabled := EnableMenu;
end; { End TFormSource.UpdateMenuItems.                                        }


{------------------------------------------------------------------------------}
{ File Menu.                                                                   }
{------------------------------------------------------------------------------}

procedure TFormSource.FileNewClick(Sender : TObject);
begin
  pBmp     := Nil;
  pBmpInfo := Nil;
  Bitmap.Free;
  Bitmap := TBitmap.Create;
  Bitmap.OnChange := ImageChanged;

  if IsDeviceOK
  then begin
       IsDirty   := False;
       IsNewFile := True;
       IsSaved   := True;
       FileSave.Enabled := False;

       {---- Clear image from Device. ----}
       { Insert Your code.                }

       Refresh;
  end;
  UpdateMenuItems(Sender);
end; { End TFormSource.FileNewClick.                                           }


procedure TFormSource.FileOpenClick(Sender : TObject);
begin
  OpenDialog.Filter   := 'Bitmap|*.BMP';
  {$IFNDEF VER90}
    OpenDialog.Filter := OpenDialog.Filter + '|JPEG|*.JPG';
  {$ENDIF}
  {$IFDEF INCTIFF}
    OpenDialog.Filter := OpenDialog.Filter + '|TIFF 5.0|*.TIF';
  {$ENDIF}
  case OpenDialog.FilterIndex of
  1 : OpenDialog.FileName := '*.BMP';
  2 : OpenDialog.FileName := '*.JPG';
  3 : OpenDialog.FileName := '*.TIF';
  end;
  if OpenDialog.Execute
  then begin
       { Clean up, remove previously loaded bitmaps.                           }
       Bitmap.Free;
       Bitmap := TBitmap.Create;
       Bitmap.OnChange := ImageChanged;

       if (Pos('.BMP', Uppercase(OpenDialog.FileName)) <> 0)
       then Bitmap.LoadFromFile(OpenDialog.FileName);
       {$IFNDEF VER90}
         if (Pos('.JPG', Uppercase(OpenDialog.FileName)) <> 0)
         then begin
              JPEGImage := TJPEGImage.Create;
              JPEGImage.Performance := jpBestQuality;
              JPEGImage.LoadFromFile(OpenDialog.FileName);
              Bitmap.Width  := JPEGImage.Width;
              Bitmap.Height := JPEGImage.Height;
              Bitmap.Canvas.Draw(0, 0, JPEGImage);
              JPEGImage.Free;
         end;
       {$ENDIF}
       {$IFDEF INCTIFF}
         if (Pos('.TIF', Uppercase(OpenDialog.FileName)) <> 0)
         then LoadTIFFToBitmap(PChar(Lowercase(OpenDialog.FileName)), Bitmap);
       {$ENDIF}

       ClientWidth   := Bitmap.Width;
       ClientHeight  := Bitmap.Height;
       Refresh;
  end;
  UpdateMenuItems(Sender);
end; { End TFormSource.FileOpenClick.                                          }


procedure TFormSource.FileSaveClick(Sender : TObject);
begin
  SaveDialog.Filter   := 'Bitmap|*.BMP';
  {$IFNDEF VER90}
    SaveDialog.Filter := SaveDialog.Filter + '|JPEG|*.JPG';
  {$ENDIF}
  {$IFDEF INCTIFF}
    SaveDialog.Filter := SaveDialog.Filter + '|TIFF 5.0|*.TIF';
  {$ENDIF}
  case SaveDialog.FilterIndex of
  1 : SaveDialog.FileName := '*.BMP';
  2 : SaveDialog.FileName := '*.JPG';
  3 : SaveDialog.FileName := '*.TIF';
  end;
  if (Pos('.BMP', SaveDialog.FileName) <> 0)
  then begin
       SaveDialog.FilterIndex := 1;
       SaveDialog.FileName := '*.BMP';
  end;
  if (Pos('.JPG', SaveDialog.FileName) <> 0)
  then begin
       SaveDialog.FilterIndex := 2;
       SaveDialog.FileName := '*.JPG';
  end;
  if (Pos('.TIF', SaveDialog.FileName) <> 0)
  then begin
       SaveDialog.FilterIndex := 3;
       SaveDialog.FileName := '*.TIF';
  end;

  if SaveDialog.Execute
  then begin
       if (Pos('.', SaveDialog.FileName) > 0)
       then SaveDialog.FileName := Copy(SaveDialog.FileName, 1, Pos('.', SaveDialog.FileName) - 1);
       case SaveDialog.FilterIndex of
       1 : SaveDialog.FileName := SaveDialog.FileName + '.BMP';
       2 : SaveDialog.FileName := SaveDialog.FileName + '.JPG';
       3 : SaveDialog.FileName := SaveDialog.FileName + '.TIF';
       end;
       if (Pos('.BMP', Uppercase(SaveDialog.FileName)) <> 0)
       then Bitmap.SaveToFile(SaveDialog.FileName);
       {$IFNDEF VER90}
         if (Pos('.JPG', Uppercase(SaveDialog.FileName)) <> 0)
         then begin
              JPEGImage := TJPEGImage.Create;
              JPEGImage.Assign(Bitmap);
              if (Bitmap.PixelFormat <> pf24bit)
              then JPEGImage.PixelFormat := jf24bit;
              JPEGImage.Scale := jsFullSize;
              { Dont compress JPEG image, ie. set quality to 100.           }
              JPEGImage.CompressionQuality  := 100;
              JPEGImage.ProgressiveEncoding := True;
              JPEGImage.SaveToFile(SaveDialog.FileName);
              JPEGImage.Free;
         end;
       {$ENDIF}
       {$IFDEF INCTIFF}
         if (Pos('.TIF', SaveDialog.FileName) <> 0)
         then SaveTIFFToBitmap(PChar(SaveDialog.FileName), Bitmap);
       {$ENDIF}
  end;
end; { End TFormSource.FileSaveClick.                                          }


procedure TFormSource.FileSnapClick(Sender : TObject);
begin
  { Clean up, remove previously loaded bitmaps.                                }
  pBmp     := Nil;
  pBmpInfo := Nil;
  Bitmap.Free;
  Bitmap := TBitmap.Create;
  Bitmap.OnChange := ImageChanged;

  { Insert your code, that assigns an image to Bitmap.                         }

  MessageBox(Handle, 'This function is Not implemented.', 'TWDSCR32', MB_OK);
end; { End TFormSource.FileSnapClick.                                          }


procedure TFormSource.FileGrapClick(Sender : TObject);
begin
  { Clean up, remove previously loaded bitmaps.                                }
  pBmp     := Nil;
  pBmpInfo := Nil;
  Bitmap.Free;
  Bitmap := TBitmap.Create;
  Bitmap.OnChange := ImageChanged;

  { Insert your code, that assigns an image to Bitmap.                         }

  MessageBox(Handle, 'This function is Not implemented.', 'TWDSCR32', MB_OK);
end; { End TFormSource.FileGrapClick.                                          }


procedure TFormSource.PrintItemClick(Sender : TObject);
begin
  if (Bitmap <> Nil)
  then begin
       if Not(Bitmap.Empty)
       then begin
            if PrintDialog.Execute
            then begin
                 FormSource.PrintScale := POPRINTTOFIT;
                 FormSource.Print;
            end;
       end
       else ShowMessage('There is nothing to print!');
  end;

  if PrintDialog.Execute
  then begin
       Printer.BeginDoc;
       Printer.Canvas.TextOut(100, 100, 'Print of picture acquired by TWAIN Driver');
       Printer.Canvas.Draw(100, 150, Bitmap);
       Printer.EndDoc;
  end;
end; { End TFormSource.PrintClick.                                             }


procedure TFormSource.SetupPrinterClick(Sender : TObject);
begin
  SetupDialog.Execute;
end; { End TFormSource.SetupPrinterClick.                                      }


procedure TFormSource.FileExitClick(Sender : TObject);
var ch  : char;
    F   : File;
begin
  { ask the application to close us ??                                         }

  {$IFDEF DEBUG}
  {
    MessageBox(Handle, 'Too late, child suicide.  Exit message.',
                       'TW_EXIT, Source Debug', MB_OK);
  }
  {$ENDIF}

  if IsDeviceOK
  then begin
  {---- Stop image acquisition.                                                }
  { Insert Your code.                                                          }

       if Bitmap.Empty
       then IsNewFile := True
       else IsNewFile := False;
  end
  else IsNewFile := True;

  if IsNewFile
  then NotifyCloseDSReq
  { Ask the application to close us ??                                         }
  { Send message back to the application                                       }
  { that we are ready                                                          }
  else begin
       { Create temporary file name. Needed for file transfer.                 }
       Ch := 'C';
       GetTempPath(SizeOf(TWAINFileName) - 1, TWAINFileName);
       GetTempFileName(TWAINFileName, '', word(0), TWAINFileName);
       EraseFile(TWAINFileName);

       { Notify receiving app, that transfer is ready.                         }
       NotifyXferReady;
  end;
end; { End TFormSource.FileExitClick.                                          }


{------------------------------------------------------------------------------}
{ Image Menu.                                                                  }
{------------------------------------------------------------------------------}

procedure TFormSource.GainOffsetClick(Sender : TObject);
begin
  { Some function to enhance contrast in your image !                          }

  MessageBox(Handle, 'This function is Not implemented.', 'TWDSCR32', MB_OK);
end; { End TFormSource.GainOffsetClick.                                        }


procedure TFormSource.InvertClick(Sender : TObject);
begin
  { Some function to invert your image !                                       }

  MessageBox(Handle, 'This function is Not implemented.', 'TWDSCR32', MB_OK);
end; { End TFormSource.InvertClick.                                            }


procedure TFormSource.ResolutionClick(Sender : TObject);
begin
  { function to enable the user to change resolution of the image.             }

  MessageBox(Handle, 'This function is Not implemented.', 'TWDSCR32', MB_OK);
end; { End TFormSource.ResolutionClick.                                        }


{------------------------------------------------------------------------------}
{ Help Menu.                                                                   }
{------------------------------------------------------------------------------}

procedure TFormSource.AboutClick(Sender : TObject);
begin
  { The About Box.                                                             }
  FormAbout := TFormAbout.Create(Application);
  FormAbout.ShowModal;
  FormAbout.Close
end; { End TFormSource.AboutClick.                                             }


function TFormSource.CopyBmpToStream : boolean;
var pBmpFileHeader : PBitmapFileHeader;
    LongWidth      : longint;
    DestRect       : TRect;
begin
  { Make a copy of the bitmap (DIB) to a memory stream.                        }
  { Ensure that you get hold of the bitmap info & data.                        }
  { Oddly, Borland does not allow you to get this directly from the            }
  { TBitmap.Handle !                                                           }

  Result   := False;
  pBmp     := Nil;
  pBmpInfo := Nil;
  if Assigned(Bitmap) and
     Assigned(ImageStream)
  then begin
       if Not(Bitmap.Empty)
       then begin
            { If the image is actually a JPEG compressed image loaded into the }
            { Bitmap, then a call to: oh - extra tricks need be done !?!       }
            { Actually the image would then have been loaded into a Picture -> }
            { the call is then:                                                }
            { TJPEG(ThePicture.Picture).DIBNeeded;                             }
            { and ThePicture is of type TImage.                                }
            { The ThePicture.Picture.Graphics will afterwoods have to be       }
            { assigned to the Bitmap.                                          }


            { Check ImagePixFrame (in twd_prot) for position/size of image to  }
            { return.                                                          }
            { Pixel, Inch, centimeter are all treated as pixels.               }
            { The image before entering in Bitmap should be croped according to}
            { ImagePixFrame.                                                   }
            { Check if application did set Image Layout.                       }
            if (ImagePixFrame.Left > 0)  or (ImagePixFrame.Top > 0) or
               (ImagePixFrame.Right > 0) or (ImagePixFrame.Bottom > 0)
            then begin { Crop image to rectangle given by ImageLayout.Frame.   }
                 if (ImagePixFrame.Left > Bitmap.Width)
                 then ImagePixFrame.Left := Bitmap.Width;
                 if (ImagePixFrame.Top > Bitmap.Height)
                 then ImagePixFrame.Top := Bitmap.Height;
                 if (ImagePixFrame.Right > Bitmap.Width)
                 then ImagePixFrame.Right := Bitmap.Width;
                 if (ImagePixFrame.Bottom > Bitmap.Height)
                 then ImagePixFrame.Bottom := Bitmap.Height;

                 DestRect.Left   := 0;
                 DestRect.Top    := 0;
                 DestRect.Right  := ImagePixFrame.Right - ImagePixFrame.Left;
                 DestRect.Bottom := ImagePixFrame.Bottom - ImagePixFrame.Top;

                 Bitmap.Canvas.CopyRect(DestRect, Bitmap.Canvas, ImagePixFrame);
                 Bitmap.Width  := DestRect.Right;
                 Bitmap.Height := DestRect.Bottom;
            end;

            Bitmap.HandleType := bmDIB;
            Bitmap.SaveToStream(ImageStream); { Get DIB copy in memory.        }
            ImageStream.Position := 0;
            pBmpFileHeader := ImageStream.Memory;
            pBmpInfo       := PBitmapInfo(@PChar(pBmpFileHeader)[SizeOf(TBitmapFileHeader)]);

            with pBmpInfo^.bmiHeader
            do begin
               { Compensate for strange Delphi behaviour with certain bitmaps !}
               biSize         := 40;
               biWidth        := Bitmap.Width;
               biHeight       := Bitmap.Height;
               biPlanes       := 1;
               case Bitmap.PixelFormat of
               pf1bit  : biBitCount := 1;
               pf4bit  : biBitCount := 4;
               pf8bit  : biBitCount := 8;
               pf15bit : biBitCount := 15;
               pf16bit : biBitCount := 16;
               pf24bit : biBitCount := 24;
               pf32bit : biBitCount := 32;
               end;
               LongWidth := (((biWidth * biBitCount) + 31) div 32) * 4;
               biSizeImage    := LongWidth * biHeight;
               biClrUsed      := 0;
               biClrImportant := 0;
               biCompression  := BI_RGB;

               { Calculate resolution in pixels per meter.                     }
               case UnitType[UnitEnum.CurrentIndex] of
               TWUN_INCHES      : begin
                                    biXPelsPerMeter := Round(100 * FIX32ToFloat(TW_FIX32(xRes.CurrentValue)) / 2.54);
                                    biYPelsPerMeter := Round(100 * FIX32ToFloat(TW_FIX32(yRes.CurrentValue)) / 2.54);
                                  end;
               TWUN_CENTIMETERS : begin
                                    biXPelsPerMeter := Round(100 * FIX32ToFloat(TW_FIX32(xRes.CurrentValue)));
                                    biYPelsPerMeter := Round(100 * FIX32ToFloat(TW_FIX32(yRes.CurrentValue)));
                                  end;
               TWUN_PICAS,
               TWUN_POINTS,
               TWUN_PIXELS      : begin { Set dpi to 300.                      }
                                    biXPelsPerMeter := Round(100 * 300 / 2.54);
                                    biYPelsPerMeter := Round(100 * 300 / 2.54);
                                  end;
               end;
            end;
            pBmp           := @PChar(pBmpFileHeader)[pBmpFileHeader^.bfoffBits];
            Result         := True;
       end;
  end;
end; { End TFormSource.CopyBmpToStream.                                        }


procedure TFormSource.ImageChanged(Sender : TObject);
begin
  { This function is called whenever the content of Bitmap is changed.         }
  pBmp     := Nil;
  pBmpInfo := Nil;
  ImageStream.Clear;
end; { End TFormSource.ImageChanged.                                           }


end. { End Unit USRCTWN.                                                       }

